/*
 *      lark-app.c
 *
 *      Copyright 2016-2018 Alex <alex@linuxonly.ru>
 *
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include "lark-app.h"
#include "lark-main-window.h"
#include "lark-pref-dialog.h"
#include "lark-gstreamer.h"
#include "lark-rhvoice.h"
#include "lark-espeak.h"
#include "lark-core.h"
#include "lark-document-loader.h"

#include "../resources.h"

struct _LarkApp
{
	GtkApplication parent_instance;
	/*Objects*/
	LarkRHVoice *rhvoice;
	LarkEspeak *espeak;
	LarkGstreamer *gstreamer;
	LarkCore *core;
	LarkDocumentLoader *docloader;
	/*Errors*/
	GError *gsterror;
	/*Settings*/
	GSettings *prefsettings;
	GSettings *eqsettings;
	GSettings *expsettings;
	gchar **filenames;
	/*Widgets*/
	GtkWidget *mainwindow;
};

struct _LarkAppClass
{
	GtkApplicationClass parent_class;
};

G_DEFINE_TYPE (LarkApp, lark_app, GTK_TYPE_APPLICATION)

enum {
	PROP_0,
	PROP_FILENAMES,
	LAST_PROP
};

static GParamSpec *properties [LAST_PROP];

static void lark_app_get_property(GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
	LarkApp *self = (LarkApp *)object;

	switch (prop_id) {
		case PROP_FILENAMES:
			g_value_set_pointer(value, self->filenames);
			break;
	}
}

static void lark_app_set_property(GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
	LarkApp *self = (LarkApp *)object;

	switch (prop_id) {
		case PROP_FILENAMES:
			self->filenames = g_value_get_pointer(value);
			break;
	}
}

static void lark_app_preferences_activated(GSimpleAction *action, GVariant *parameter, gpointer app)
{
	LarkApp *self = LARK_APP(app);
	GtkWidget *prefdialog;

	prefdialog = lark_pref_dialog_new(GTK_WINDOW(self->mainwindow), self->core, LARK_TTS_ENGINE(self->rhvoice), LARK_TTS_ENGINE(self->espeak), self->prefsettings);
	gtk_window_present(GTK_WINDOW(prefdialog));
}

static void lark_app_about_activated (GSimpleAction *action, GVariant *parameter, gpointer app)
{
	GdkPixbuf *logo;
	GError *error;
	gchar *artists[] = {
		"rg1024 <https://openclipart.org/user-detail/rg1024>",
		NULL
	};
	gchar *authors[] = {
		"Alex <alex@linuxonly.ru>",
		NULL
	};
	gchar *documenters[] = {
		"Alex <alex@linuxonly.ru>",
		NULL
	};

	LarkApp *self = LARK_APP(app);

	error = NULL;

	logo = gdk_pixbuf_new_from_file_at_scale(RESOURCE_LOGO_FILE, 128, 128, FALSE, &error);

	if ((logo == NULL) && (error != NULL)) {
		printf("Error loading pixbuf: %s\n", error->message);
	}

	gtk_show_about_dialog(GTK_WINDOW(self->mainwindow),
						"artists",            artists,
						"authors",            authors,
						"comments",           _("Simple audiobook creator with good Russian and English languages support based on RHVoice and ESpeak TTS engines"),
						"copyright",          _("Copyright 2016 - 2018, Alex"),
						"documenters",        documenters,
						"license-type",       GTK_LICENSE_GPL_3_0,
						"logo",               logo,
						"program-name",       _("Lark Audiobook Creator"),
						"translator-credits", _("Alex <alex@linuxonly.ru>"),
						"version",            RESOURCE_VERSION,
						"website",            "https://linuxonly.ru/page/lark-audiobook-creator",
						"website-label",      _("Homepage"),
						NULL);

	if (logo != NULL) {
		g_object_unref(logo);
	}
}

static void lark_app_quit_activated(GSimpleAction *action, GVariant *parameter, gpointer app)
{
	g_application_quit(G_APPLICATION(app));
}

static GActionEntry lark_app_entries[] =
{
	{ "preferences", lark_app_preferences_activated, NULL, NULL, NULL },
	{ "about",       lark_app_about_activated,       NULL, NULL, NULL },
	{ "quit",        lark_app_quit_activated,        NULL, NULL, NULL }
};

static void lark_app_startup(GApplication *app)
{
	GtkBuilder *builder;
	GMenuModel *app_menu;

	G_APPLICATION_CLASS(lark_app_parent_class)->startup(app);

	g_action_map_add_action_entries(G_ACTION_MAP(app), lark_app_entries, G_N_ELEMENTS(lark_app_entries), app);

	builder = gtk_builder_new_from_resource("/ru/linuxonly/lark/ui/app-menu.ui");
	app_menu = G_MENU_MODEL(gtk_builder_get_object(builder, "app-menu"));
	gtk_application_set_app_menu(GTK_APPLICATION(app), app_menu);
	g_object_unref(builder);
}

static void lark_app_activate(GApplication *app)
{
	GtkWidget *dialog;

	LarkApp *self = LARK_APP(app);

	/*Show message it there are no needed plugins in system*/
	if (self->gstreamer == NULL) {
		if (self->gsterror != NULL) {
			dialog = gtk_message_dialog_new(NULL,
											GTK_DIALOG_USE_HEADER_BAR,
											GTK_MESSAGE_ERROR,
											GTK_BUTTONS_CLOSE,
											"%s",
											self->gsterror->message);
			gtk_dialog_run(GTK_DIALOG(dialog));
			gtk_widget_destroy(dialog);

			g_error_free(self->gsterror);
			g_application_quit(G_APPLICATION(app));
		}
	}
	/*Create new window if not exists*/
	if (self->mainwindow == NULL) {
		self->mainwindow = lark_main_window_new(LARK_APP(app), self->gstreamer, self->core, self->docloader, self->prefsettings, self->eqsettings, self->expsettings, self->filenames);
	}
	/*Move main window to front*/
	gtk_window_present(GTK_WINDOW(self->mainwindow));
}

static void lark_app_shutdown(GApplication *object)
{
	LarkApp *self = (LarkApp *)object;

	g_object_unref(self->core);
	g_object_unref(self->docloader);
	g_object_unref(self->espeak);
	g_object_unref(self->rhvoice);
	g_object_unref(self->gstreamer);

	G_APPLICATION_CLASS(lark_app_parent_class)->shutdown(object);
}

static void lark_app_class_init(LarkAppClass *klass)
{
	GApplicationClass *app_class = G_APPLICATION_CLASS(klass);
	GObjectClass *object_class = G_OBJECT_CLASS(klass);

	app_class->startup  = lark_app_startup;
	app_class->activate = lark_app_activate;
	app_class->shutdown = lark_app_shutdown;
	object_class->get_property = lark_app_get_property;
	object_class->set_property = lark_app_set_property;

	properties[PROP_FILENAMES] =
		g_param_spec_pointer("filenames",
							"Pointer to filenames array",
							"Pointer to filenames array",
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);

	g_object_class_install_properties(object_class, LAST_PROP, properties);
}

static void lark_app_init(LarkApp *self)
{
	/*Settings*/
	self->prefsettings = g_settings_new("ru.linuxonly.lark.preferences");
	self->eqsettings = g_settings_new("ru.linuxonly.lark.equalizer");
	self->expsettings = g_settings_new("ru.linuxonly.lark.export");

	/*Objects*/
	/*GStreamer*/
	self->gsterror = NULL;
	self->gstreamer = lark_gstreamer_new(&self->gsterror);
	/*TTS engines*/
	self->rhvoice = lark_rhvoice_new();
	self->espeak = lark_espeak_new();
	/*Document loader*/
	self->docloader = lark_document_loader_new();
	/*App core*/
	self->core = lark_core_new(self->gstreamer, self->rhvoice, self->espeak, self->prefsettings, self->eqsettings, self->expsettings);
}

LarkApp *lark_app_new(gchar **filenames)
{
	LarkApp *larkapp;

	larkapp = g_object_new(lark_app_get_type(), "application-id", "ru.linuxonly.lark", "register-session", TRUE, "flags", G_APPLICATION_FLAGS_NONE, "filenames", filenames, NULL);

	return larkapp;
}
