/*
 *      lark-gstreamer.c
 *
 *      Copyright 2016-2018 Alex <alex@linuxonly.ru>
 *
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <gst/gst.h>
#include <gst/audio/audio.h>
#include <gst/app/gstappsrc.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <glib/gstdio.h>

#include "lark-gstreamer.h"

#define LARK_QUEUE_HIGH_LEVEL_BYTES 10*1024*1024
#define LARK_QUEUE_LOW_LEVEL_BYTES             0

static GstPadProbeReturn lark_gstreamer_inspect_output_buffer(GstPad *pad, GstPadProbeInfo *info, LarkGstreamer *self);

#define LARK_GSTREAMER_ERROR  g_quark_from_static_string("ru.linuxonly.lark.gstreamer.error")

struct _LarkGstreamer
{
	GObject parent_instance;
	/*Objects*/
	
	/*Data*/
	GstPipeline *pipeline;
	GstElement *appsrc;
	GstElement *converter;
	GstElement *equalizer;
	GstElement *audiosink;
	GstElement *encoder;
	GstElement *muxer;
	GstElement *filesink;
	GstElement *oldmuxer;
	GstElement *oldfilesink;
	
	gboolean vorbisavailable;
	gboolean opusavailable;
	gboolean flacavailable;
	gboolean mp3available;
	gboolean aacavailable;
	
	guint format;
	guint samplerate;
	guint channels;
	guint target;
	guint quality;
	gchar *path;
	gchar *filename;
	
	guint limitmode;
	guint sizelimit;
	guint timelimit;
	
	guint queuebytes;
	
	gchar *titletag;
	gchar *authortag;
	gchar *commenttag;
	/*gchar *datetag;*/
	
	gfloat eq100gain;
	gfloat eq300gain;
	gfloat eq1kgain;
	gfloat eq3kgain;
	gfloat eq8kgain;
	
	GstTagList *taglist;
	
	guint numsamples;
	guint numfragments;
	gsize fullsize;
	gsize filesize;
	GstClockTime fullduration;
	GstClockTime fileduration;
};

struct _LarkGstreamerClass
{
	GObjectClass parent_class;
};

static void lark_gstreamer_initable_iface_init (GInitableIface *initable_iface);

G_DEFINE_TYPE_WITH_CODE (LarkGstreamer, lark_gstreamer, G_TYPE_OBJECT,
                         G_IMPLEMENT_INTERFACE (G_TYPE_INITABLE, lark_gstreamer_initable_iface_init ));

enum {
	PROP_0,
	PROP_QUEUE_BYTES,
	PROP_VORBIS_AVAILABLE,
	PROP_OPUS_AVAILABLE,
	PROP_FLAC_AVAILABLE,
	PROP_MP3_AVAILABLE,
	PROP_AAC_AVAILABLE,
	PROP_FORMAT,
	PROP_SAMPLERATE,
	PROP_CHANNELS,
	PROP_TARGET,
	PROP_QUALITY,
	PROP_PATH,
	PROP_FILENAME,
	PROP_LIMITMODE,
	PROP_SIZELIMIT,
	PROP_TIMELIMIT,
	PROP_TITLE_TAG,
	PROP_AUTHOR_TAG,
	PROP_COMMENT_TAG,
	/*PROP_DATE_TAG,*/
	LAST_PROP
};

static GParamSpec *properties [LAST_PROP];

enum {
	SIGNAL_FINISHED,
	LAST_SIGNAL
};

static guint signals[LAST_SIGNAL];

static void lark_gstreamer_get_property(GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
	LarkGstreamer *self = (LarkGstreamer *)object;
	
	switch (prop_id) {
		case PROP_QUEUE_BYTES:
			g_value_set_uint(value, self->queuebytes);
			break;
		case PROP_VORBIS_AVAILABLE:
			g_value_set_boolean(value, self->vorbisavailable);
			break;
		case PROP_OPUS_AVAILABLE:
			g_value_set_boolean(value, self->opusavailable);
			break;
		case PROP_FLAC_AVAILABLE:
			g_value_set_boolean(value, self->flacavailable);
			break;
		case PROP_MP3_AVAILABLE:
			g_value_set_boolean(value, self->mp3available);
			break;
		case PROP_AAC_AVAILABLE:
			g_value_set_boolean(value, self->aacavailable);
			break;
		case PROP_FORMAT:
			g_value_set_uint(value, self->format);
			break;
		case PROP_SAMPLERATE:
			g_value_set_uint(value, self->samplerate);
			break;
		case PROP_CHANNELS:
			g_value_set_uint(value, self->channels);
			break;
		case PROP_TARGET:
			g_value_set_uint(value, self->target);
			break;
		case PROP_QUALITY:
			g_value_set_uint(value, self->quality);
			break;
		case PROP_PATH:
			g_value_set_string(value, self->path);
			break;
		case PROP_FILENAME:
			g_value_set_string(value, self->filename);
			break;
		case PROP_LIMITMODE:
			g_value_set_uint(value, self->limitmode);
			break;
		case PROP_SIZELIMIT:
			g_value_set_uint(value, self->sizelimit);
			break;
		case PROP_TIMELIMIT:
			g_value_set_uint(value, self->timelimit);
			break;
		case PROP_TITLE_TAG:
			g_value_set_string(value, self->titletag);
			break;
		case PROP_AUTHOR_TAG:
			g_value_set_string(value, self->authortag);
			break;
		case PROP_COMMENT_TAG:
			g_value_set_string(value, self->commenttag);
			break;
		/*case PROP_DATE_TAG:
			g_value_set_string(value, self->datetag);
			break;*/
	}
}

static void lark_gstreamer_set_property(GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
	LarkGstreamer *self = (LarkGstreamer *)object;
	
	switch (prop_id) {
		case PROP_QUEUE_BYTES:
			self->queuebytes = g_value_get_uint(value);
			break;
		case PROP_FORMAT:
			self->format = g_value_get_uint(value);
			break;
		case PROP_SAMPLERATE:
			self->samplerate = g_value_get_uint(value);
			break;
		case PROP_CHANNELS:
			self->channels = g_value_get_uint(value);
			break;
		case PROP_TARGET:
			self->target = g_value_get_uint(value);
			break;
		case PROP_QUALITY:
			self->quality = g_value_get_uint(value);
			break;
		case PROP_PATH:
			if (self->path != NULL) {
				g_free(self->path);
			}
			self->path = g_strdup(g_value_get_string(value));
			break;
		case PROP_FILENAME:
			if (self->filename != NULL) {
				g_free(self->filename);
			}
			self->filename = g_strdup(g_value_get_string(value));
			break;
		case PROP_LIMITMODE:
			self->limitmode = g_value_get_uint(value);
			break;
		case PROP_SIZELIMIT:
			self->sizelimit = g_value_get_uint(value);
			break;
		case PROP_TIMELIMIT:
			self->timelimit = g_value_get_uint(value);
			break;
		case PROP_TITLE_TAG:
			if (self->titletag != NULL) {
				g_free(self->titletag);
			}
			self->titletag = g_strdup(g_value_get_string(value));
			break;
		case PROP_AUTHOR_TAG:
			if (self->authortag != NULL) {
				g_free(self->authortag);
			}
			self->authortag = g_strdup(g_value_get_string(value));
			break;
		case PROP_COMMENT_TAG:
			if (self->commenttag != NULL) {
				g_free(self->commenttag);
			}
			self->commenttag = g_strdup(g_value_get_string(value));
			break;
		/*case PROP_DATE_TAG:
			if (self->datetag != NULL) {
				g_free(self->datetag);
			}
			self->datetag = g_strdup(g_value_get_string(value));
			break;*/
	}
}

static void lark_gstreamer_constructed(GObject *object)
{
	LarkGstreamer *self __attribute__((unused)) = (LarkGstreamer *)object;
	
	G_OBJECT_CLASS(lark_gstreamer_parent_class)->constructed(object);
}

static void lark_gstreamer_finalize(GObject *object)
{
	LarkGstreamer *self = (LarkGstreamer *)object;
	
	if (self->pipeline != NULL) {
		gst_element_set_state(GST_ELEMENT(self->pipeline), GST_STATE_NULL);
	}
	
	G_OBJECT_CLASS(lark_gstreamer_parent_class)->finalize(object);
}

static void lark_gstreamer_class_init(LarkGstreamerClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS(klass);
	
	object_class->get_property = lark_gstreamer_get_property;
	object_class->set_property = lark_gstreamer_set_property;
	
	object_class->constructed = lark_gstreamer_constructed;
	object_class->finalize = lark_gstreamer_finalize;
	
	properties[PROP_QUEUE_BYTES] =
		g_param_spec_uint("queue-bytes",
							"Appsource queue length",
							"Appsource queue length in bytes",
							LARK_QUEUE_LOW_LEVEL_BYTES,
							LARK_QUEUE_HIGH_LEVEL_BYTES,
							LARK_QUEUE_LOW_LEVEL_BYTES,
							G_PARAM_READWRITE);
	
	properties[PROP_VORBIS_AVAILABLE] =
		g_param_spec_boolean("vorbis-available",
							"Vorbis encoder availability flag",
							"Vorbis encoder availability flag",
							FALSE,
							G_PARAM_READABLE);
	
	properties[PROP_OPUS_AVAILABLE] =
		g_param_spec_boolean("opus-available",
							"Opus encoder availability flag",
							"Opus encoder availability flag",
							FALSE,
							G_PARAM_READABLE);
	
	properties[PROP_FLAC_AVAILABLE] =
		g_param_spec_boolean("flac-available",
							"FLAC encoder availability flag",
							"FLAC encoder availability flag",
							FALSE,
							G_PARAM_READABLE);
							
	properties[PROP_MP3_AVAILABLE] =
		g_param_spec_boolean("mp3-available",
							"MP3 encoder availability flag",
							"MP3 encoder availability flag",
							FALSE,
							G_PARAM_READABLE);
							
	properties[PROP_AAC_AVAILABLE] =
		g_param_spec_boolean("aac-available",
							"AAC encoder availability flag",
							"AAC encoder availability flag",
							FALSE,
							G_PARAM_READABLE);
	
	properties[PROP_FORMAT] =
		g_param_spec_uint("format",
							"Wave data format",
							"Wave data format",
							LARK_GSTREAMER_SRC_FORMAT_U8,
							LARK_GSTREAMER_SRC_FORMAT_S32BE,
							LARK_GSTREAMER_SRC_FORMAT_S16LE,
							G_PARAM_READWRITE);
							
	properties[PROP_SAMPLERATE] =
		g_param_spec_uint("sample-rate",
							"Wave data sample rate",
							"Wave data sample rate",
							LARK_GSTREAMER_MIN_SAMPERATE,
							LARK_GSTREAMER_MAX_SAMPERATE,
							LARK_GSTREAMER_DEF_SAMPERATE,
							G_PARAM_READWRITE);
							
	properties[PROP_CHANNELS] =
		g_param_spec_uint("channels",
							"Wave channels number",
							"Wave channels number",
							LARK_GSTREAMER_MIN_CHANNELS,
							LARK_GSTREAMER_MAX_CHANNELS,
							LARK_GSTREAMER_DEF_CHANNELS,
							G_PARAM_READWRITE);
							
	properties[PROP_TARGET] =
		g_param_spec_uint("target",
							"Wave output target",
							"Wave output target",
							LARK_GSTREAMER_TARGET_AUDIO,
							LARK_GSTREAMER_TARGET_AAC,
							LARK_GSTREAMER_TARGET_AUDIO,
							G_PARAM_READWRITE);
							
	properties[PROP_QUALITY] =
		g_param_spec_uint("quality",
							"Wave transcoding quality",
							"Wave transcoding quality",
							LARK_GSTREAMER_MIN_QUALITY,
							LARK_GSTREAMER_MAX_QUALITY,
							LARK_GSTREAMER_DEF_QUALITY,
							G_PARAM_READWRITE);
	
	properties[PROP_PATH] =
		g_param_spec_string("path",
							"Directory to store output file",
							"Directory to store output file",
							"/tmp",
							G_PARAM_READWRITE);
							
	properties[PROP_FILENAME] =
		g_param_spec_string("filename",
							"Output file name",
							"Output file name",
							"audiobook",
							G_PARAM_READWRITE);
							
	properties[PROP_LIMITMODE] =
		g_param_spec_uint("limit-mode",
							"Limit mode",
							"Limit mode",
							LARK_GSTREAMER_LIMIT_MODE_NONE,
							LARK_GSTREAMER_LIMIT_MODE_TIME,
							LARK_GSTREAMER_LIMIT_MODE_NONE,
							G_PARAM_READWRITE);
	
	properties[PROP_SIZELIMIT] =
		g_param_spec_uint("size-limit",
							"Output file size limit in bytes",
							"Output file size limit in bytes",
							LARK_GSTREAMER_MIN_SIZE_LIMIT,
							LARK_GSTREAMER_MAX_SIZE_LIMIT,
							LARK_GSTREAMER_DEF_SIZE_LIMIT,
							G_PARAM_READWRITE);
							
	properties[PROP_TIMELIMIT] =
		g_param_spec_uint("time-limit",
							"Output file time limit in seconds",
							"Output file time limit in seconds",
							LARK_GSTREAMER_MIN_TIME_LIMIT,
							LARK_GSTREAMER_MAX_TIME_LIMIT,
							LARK_GSTREAMER_DEF_TIME_LIMIT,
							G_PARAM_READWRITE);
							
	
	properties[PROP_TITLE_TAG] =
		g_param_spec_string("title-tag",
							"Title tag",
							"Title tag",
							"",
							G_PARAM_READWRITE);
							
	properties[PROP_AUTHOR_TAG] =
		g_param_spec_string("author-tag",
							"Author tag",
							"Author tag",
							"",
							G_PARAM_READWRITE);
							
	properties[PROP_COMMENT_TAG] =
		g_param_spec_string("comment-tag",
							"Comment tag",
							"Comment tag",
							"",
							G_PARAM_READWRITE);
							
	/*properties[PROP_DATE_TAG] =
		g_param_spec_string("date-tag",
							"Date tag",
							"Date tag",
							"",
							G_PARAM_READWRITE);*/
							
	g_object_class_install_properties (object_class, LAST_PROP, properties);
	
	signals[SIGNAL_FINISHED] =
		g_signal_new("finished",
							G_TYPE_FROM_CLASS(klass),
							G_SIGNAL_RUN_FIRST,
							0,
							NULL,
							NULL,
							g_cclosure_marshal_generic,
							G_TYPE_NONE,
							2,
							G_TYPE_BOOLEAN,
							G_TYPE_ERROR);
}

static gboolean lark_gstreamer_initable_init (GInitable *initable, GCancellable *cancellable, GError **error)
{
	gint i;
	GstElementFactory *factory;
	const gchar *neddedelements[] = {"appsrc", "audioconvert", "pulsesink", "filesink", "wavenc", NULL};
		
	i = 0;
	while (neddedelements[i] != NULL) {
		factory = gst_element_factory_find(neddedelements[i]);
		if (factory == NULL) {
			if (error != NULL) {
				*error = g_error_new(LARK_GSTREAMER_ERROR, 0, _("Nedded GStreamer element \'%s\' is not found! Please install package with it first."), neddedelements[i]);
			}
			return FALSE;
		} else {
			gst_object_unref(factory);
		}
		i++;
	}
	
	return TRUE;
}

static void lark_gstreamer_initable_iface_init (GInitableIface *initable_iface)
{
	initable_iface->init = lark_gstreamer_initable_init;
}

static void lark_gstreamer_init(LarkGstreamer *self)
{
	gint i;
	GstElementFactory *factory;
	gboolean *flags[] = {&self->vorbisavailable, &self->opusavailable, &self->flacavailable, &self->mp3available, &self->aacavailable, NULL};
	const gchar *elements[][2] = {{"vorbisenc", "oggmux"}, {"opusenc", "oggmux"}, {"flacenc", NULL}, {"lamemp3enc", NULL}, {"faac", "mp4mux"}};
	
	i = 0;
	while (flags[i] != NULL) {
		*flags[i] = FALSE;
		/*Encoder*/
		factory = gst_element_factory_find(elements[i][0]);
		if (factory != NULL) {
			gst_object_unref(factory);
			if (elements[i][1] != NULL) {
				/*Encoder + muxer*/
				factory = gst_element_factory_find(elements[i][1]);
				if (factory != NULL) {
					gst_object_unref(factory);
					*flags[i] = TRUE;
				}
			} else {
				/*Only encoder*/
				*flags[i] = TRUE;
			}
		}
		i++;
	}
}

gboolean lark_gstreamer_add_data(LarkGstreamer *self, guint8 *data, gsize length)
{
	GstBuffer *buffer;
	guint64 bytes;
	
	g_return_val_if_fail(self != NULL, FALSE);
	g_return_val_if_fail(data != NULL, FALSE);
	g_return_val_if_fail(length != 0, FALSE);
	
	/*New wrapped buffer*/
	buffer = gst_buffer_new_wrapped(data, length);
	/*Timestamps*/
	GST_BUFFER_TIMESTAMP(buffer) = gst_util_uint64_scale(self->numsamples, GST_SECOND, self->samplerate);
	GST_BUFFER_DURATION(buffer) = gst_util_uint64_scale(length, GST_SECOND, self->samplerate);
	/*Get queue length*/
	g_object_get(GST_APP_SRC(self->appsrc), "current-level-bytes", &bytes, NULL);
	g_debug("Bytes in internal GStreamer queue: %" G_GUINT64_FORMAT, bytes);
	/*Push buffer*/
	gst_app_src_push_buffer(GST_APP_SRC(self->appsrc), buffer);
	/*Update sample counter*/
	switch (self->format) {
		case LARK_GSTREAMER_SRC_FORMAT_U8:
		case LARK_GSTREAMER_SRC_FORMAT_S8:
			self->numsamples += length;
			break;
		case LARK_GSTREAMER_SRC_FORMAT_U16LE:
		case LARK_GSTREAMER_SRC_FORMAT_U16BE:
		case LARK_GSTREAMER_SRC_FORMAT_S16LE:
		case LARK_GSTREAMER_SRC_FORMAT_S16BE:
			self->numsamples += length/2;
			break;
		case LARK_GSTREAMER_SRC_FORMAT_U32LE:
		case LARK_GSTREAMER_SRC_FORMAT_U32BE:
		case LARK_GSTREAMER_SRC_FORMAT_S32LE:
		case LARK_GSTREAMER_SRC_FORMAT_S32BE:
			self->numsamples += length/4;
			break;
		default:
			self->numsamples += length;
			break;
	}
	
	return TRUE;
}

gboolean lark_gstreamer_get_vorbis_available(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, FALSE);
	
	return self->vorbisavailable;	
}

gboolean lark_gstreamer_get_opus_available(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, FALSE);
	
	return self->opusavailable;	
}

gboolean lark_gstreamer_get_flac_available(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, FALSE);
	
	return self->flacavailable;	
}

gboolean lark_gstreamer_get_mp3_available(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, FALSE);
	
	return self->mp3available;	
}

gboolean lark_gstreamer_get_aac_available(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, FALSE);
	
	return self->aacavailable;	
}

void lark_gstreamer_set_format(LarkGstreamer *self, gint format)
{
	g_return_if_fail(self != NULL);
	g_return_if_fail(format >= LARK_GSTREAMER_SRC_FORMAT_U8 && format <= LARK_GSTREAMER_SRC_FORMAT_S32BE);
	
	self->format = format;
}

guint lark_gstreamer_get_format(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, LARK_GSTREAMER_SRC_FORMAT_S16LE);
	
	return self->format;
}

void lark_gstreamer_set_sample_rate(LarkGstreamer *self, guint samplerate)
{
	g_return_if_fail(self != NULL);
	g_return_if_fail(samplerate >= LARK_GSTREAMER_MIN_SAMPERATE && samplerate <= LARK_GSTREAMER_MAX_SAMPERATE);
	
	self->samplerate = samplerate;
}

guint lark_gstreamer_get_sample_rate(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, LARK_GSTREAMER_DEF_SAMPERATE);
	
	return self->samplerate;
}

void lark_gstreamer_set_channels(LarkGstreamer *self, guint channels)
{
	g_return_if_fail(self != NULL);
	g_return_if_fail(channels >= LARK_GSTREAMER_MIN_CHANNELS && channels <= LARK_GSTREAMER_MAX_CHANNELS);
	
	self->channels = channels;
}

guint lark_gstreamer_get_channels(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, LARK_GSTREAMER_DEF_CHANNELS);
	
	return self->channels;
}

void lark_gstreamer_set_target(LarkGstreamer *self, guint target)
{
	g_return_if_fail(self != NULL);
	g_return_if_fail(target >= LARK_GSTREAMER_TARGET_AUDIO && target <= LARK_GSTREAMER_TARGET_AAC);
	
	self->target = target;
}

guint lark_gstreamer_get_target(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, LARK_GSTREAMER_TARGET_AUDIO);
	
	return self->target;
}

void lark_gstreamer_set_quality(LarkGstreamer *self, guint quality)
{
	g_return_if_fail(self != NULL);
	g_return_if_fail(quality >= LARK_GSTREAMER_MIN_QUALITY && quality <= LARK_GSTREAMER_MAX_QUALITY);
	
	self->quality = quality;
}

guint lark_gstreamer_get_quality(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, LARK_GSTREAMER_DEF_QUALITY);
	
	return self->quality;
}

void lark_gstreamer_set_path(LarkGstreamer *self, gchar *path)
{
	GStatBuf buf;
	
	g_return_if_fail(self != NULL);
	g_return_if_fail(path != NULL);
	
	if ((g_stat(path, &buf) == 0) && (g_access(path, W_OK) == 0)) {
		if (S_ISDIR(buf.st_mode)) {
			if (self->path != NULL) {
				g_free(self->path);
			}
			self->path = g_strdup(path);
		}
	}
}

gchar *lark_gstreamer_get_path(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, "/tmp");
	
	return self->path;
}

void lark_gstreamer_set_file_name(LarkGstreamer *self, gchar *filename)
{
	g_return_if_fail(self != NULL);
	g_return_if_fail(filename != NULL);
	
	if (self->filename != NULL) {
		g_free(self->filename);
	}
	self->filename = g_strdup(filename);
}

gchar *lark_gstreamer_get_file_name(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, "audiobook");
	
	return self->filename;
}

void lark_gstreamer_set_limit_mode(LarkGstreamer *self, guint mode)
{
	g_return_if_fail(self != NULL);
	g_return_if_fail(mode <= LARK_GSTREAMER_LIMIT_MODE_TIME);
	
	self->limitmode = mode;
}

guint lark_gstreamer_get_limit_mode(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, LARK_GSTREAMER_LIMIT_MODE_NONE);
	
	return self->limitmode;
}

void lark_gstreamer_set_size_limit(LarkGstreamer *self, guint sizelimit)
{
	g_return_if_fail(self != NULL);
	g_return_if_fail(sizelimit >= LARK_GSTREAMER_MIN_SIZE_LIMIT && sizelimit <= LARK_GSTREAMER_MAX_SIZE_LIMIT);
		
	self->sizelimit = sizelimit;
}

guint lark_gstreamer_get_size_limit(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, 0);
	
	return self->sizelimit;
}

void lark_gstreamer_set_time_limit(LarkGstreamer *self, guint timelimit)
{
	g_return_if_fail(self != NULL);
	g_return_if_fail(timelimit >= LARK_GSTREAMER_MIN_TIME_LIMIT && timelimit <= LARK_GSTREAMER_MAX_TIME_LIMIT);
	
	self->timelimit = timelimit;
}

guint lark_gstreamer_get_time_limit(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, 0);
	
	return self->timelimit;
}

void lark_gstreamer_set_title_tag(LarkGstreamer *self, gchar *titletag)
{
	g_return_if_fail(self != NULL);
	g_return_if_fail(titletag != NULL);
	
	if (self->titletag != NULL) {
		g_free(self->titletag);
	}
	self->titletag = g_strdup(titletag);
}

gchar *lark_gstreamer_get_title_tag(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, "");
	
	return self->titletag;
}

void lark_gstreamer_set_author_tag(LarkGstreamer *self, gchar *authortag)
{
	g_return_if_fail(self != NULL);
	g_return_if_fail(authortag != NULL);
	
	if (self->authortag != NULL) {
		g_free(self->authortag);
	}
	self->authortag = g_strdup(authortag);
}

gchar *lark_gstreamer_get_author_tag(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, "");
	
	return self->authortag;
}

void lark_gstreamer_set_comment_tag(LarkGstreamer *self, gchar *commenttag)
{
	g_return_if_fail(self != NULL);
	g_return_if_fail(commenttag != NULL);
	
	if (self->commenttag != NULL) {
		g_free(self->commenttag);
	}
	self->commenttag = g_strdup(commenttag);
}

gchar *lark_gstreamer_get_comment_tag(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, "");
	
	return self->commenttag;
}

void lark_gstreamer_set_eq_gain(LarkGstreamer *self, guint freqid, gfloat gain)
{
	GstState state;
	GObject *band = NULL;
	
	g_return_if_fail(self != NULL);
	g_return_if_fail(gain >= -12.0 && gain <= 12.0);
	
	if (self->pipeline != NULL) {
		gst_element_get_state(GST_ELEMENT(self->pipeline), &state, NULL, GST_CLOCK_TIME_NONE);
	}
	
	switch (freqid) {
		case LARK_GSTREAMER_EQ_FREQ_100:
			self->eq100gain = gain;
			if ((self->pipeline != NULL) && (state > GST_STATE_PAUSED)) {
				band = gst_child_proxy_get_child_by_index(GST_CHILD_PROXY(self->equalizer), LARK_GSTREAMER_EQ_FREQ_100);
			}
			break;
		case LARK_GSTREAMER_EQ_FREQ_300:
			self->eq300gain = gain;
			if ((self->pipeline != NULL) && (state > GST_STATE_PAUSED)) {
				band = gst_child_proxy_get_child_by_index(GST_CHILD_PROXY(self->equalizer), LARK_GSTREAMER_EQ_FREQ_300);
			}
			break;
		case LARK_GSTREAMER_EQ_FREQ_1K:
			self->eq1kgain = gain;
			if ((self->pipeline != NULL) && (state > GST_STATE_PAUSED)) {
				band = gst_child_proxy_get_child_by_index(GST_CHILD_PROXY(self->equalizer), LARK_GSTREAMER_EQ_FREQ_1K);
			}
			break;
		case LARK_GSTREAMER_EQ_FREQ_3K:
			self->eq3kgain = gain;
			if ((self->pipeline != NULL) && (state > GST_STATE_PAUSED)) {
				band = gst_child_proxy_get_child_by_index(GST_CHILD_PROXY(self->equalizer), LARK_GSTREAMER_EQ_FREQ_3K);
			}
			break;
		case LARK_GSTREAMER_EQ_FREQ_8K:
			self->eq8kgain = gain;
			if ((self->pipeline != NULL) && (state > GST_STATE_PAUSED)) {
				band = gst_child_proxy_get_child_by_index(GST_CHILD_PROXY(self->equalizer), LARK_GSTREAMER_EQ_FREQ_8K);
			}
			break;
		default:
			break;
	}
	
	if (band != NULL) {
		g_object_set(G_OBJECT(band), "gain", gain, NULL);
		g_object_unref(G_OBJECT(band));
	}
}

gfloat lark_gstreamer_get_eq_gain(LarkGstreamer *self, guint freqid)
{
	gfloat gain;
	
	g_return_val_if_fail(self != NULL, 0.0);
	
	switch (freqid) {
		case LARK_GSTREAMER_EQ_FREQ_100:
			gain = self->eq100gain;
			break;
		case LARK_GSTREAMER_EQ_FREQ_300:
			gain = self->eq300gain;
			break;
		case LARK_GSTREAMER_EQ_FREQ_1K:
			gain = self->eq1kgain;
			break;
		case LARK_GSTREAMER_EQ_FREQ_3K:
			gain = self->eq3kgain;
			break;
		case LARK_GSTREAMER_EQ_FREQ_8K:
			gain = self->eq8kgain;
			break;
		default:
			gain = 0.0;
			break;
	}
	
	return gain;
}

gsize lark_gstreamer_get_full_size(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, 0);
	
	return self->fullsize + self->filesize;
}

guint64 lark_gstreamer_get_full_duration(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, 0);
	
	return (guint64)((self->fullduration + self->fileduration) / GST_SECOND);
}

void lark_gstreamer_set_queue_bytes(LarkGstreamer *self, guint queuebytes)
{
	g_return_if_fail(self != NULL);
	
	self->queuebytes = queuebytes;
}

guint lark_gstreamer_get_queue_bytes(LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, 0);
	
	return self->queuebytes;
} 

void lark_gstreamer_stop(LarkGstreamer *self, gboolean immediately)
{
	g_return_if_fail(self != NULL);
	
	if (!immediately) {
		/*Send EOS event*/
		gst_app_src_end_of_stream(GST_APP_SRC(self->appsrc));
	} else {
		/*Stop pipeline*/
		gst_tag_list_unref(self->taglist);
		gst_element_set_state(GST_ELEMENT(self->pipeline), GST_STATE_NULL);
	}
	
	g_debug("GStreamer stopped");
}

static gchar *lark_gstreamer_build_file_path(LarkGstreamer *self)
{
	gchar *filename, *fullpath, *extension;
	
	g_return_val_if_fail(self != NULL, g_strdup("/tmp/output.bin"));
	
	switch (self->target) {
		case LARK_GSTREAMER_TARGET_WAV:
			extension = ".wav";
			break;
		case LARK_GSTREAMER_TARGET_VORBIS:
		case LARK_GSTREAMER_TARGET_OPUS:
			extension = ".ogg";
			break;
		case LARK_GSTREAMER_TARGET_FLAC:
			extension = ".flac";
			break;
		case LARK_GSTREAMER_TARGET_MP3:
			extension = ".mp3";
			break;
		case LARK_GSTREAMER_TARGET_AAC:
			extension = ".aac";
			break;
		default:
			extension = ".bin";
			break;
	}
	
	if (self->limitmode != LARK_GSTREAMER_LIMIT_MODE_NONE) {
		filename = g_strdup_printf("%s (%u)%s", self->filename, self->numfragments, extension);
	} else {
		filename = g_strconcat(self->filename, extension, NULL);
	}
	
	fullpath = g_build_filename(self->path, filename, NULL);
	
	return fullpath;		
}

static gchar *lark_gstreamer_build_element_name(LarkGstreamer *self, const gchar *name)
{
	gchar *resname;
	
	g_return_val_if_fail(self != NULL, g_strdup("element"));
	g_return_val_if_fail(name != NULL, g_strdup("element"));
	
	if (self->limitmode != LARK_GSTREAMER_LIMIT_MODE_NONE) {
		resname = g_strdup_printf("%s_%u", name, self->numfragments);
	} else {
		resname = g_strdup(name);
	}
	
	return resname;	
}

static void lark_gstreamer_build_encoding_chain(LarkGstreamer *self, GstPad *srcpad)
{
	GstPad *sinkpad;
	gchar *elementname, *filepath;
	gint *iquality;
	gfloat *fquality;
	GstElement *oldencoder, *oldmuxer, *oldfilesink;
	/*Encoders quality mappings*/
	gfloat vorbisquality[] = {0.0,   0.1,   0.2,    0.3,    0.4,    0.5,    0.6,    0.7,    0.8,    0.9,    1.0};
	gint   opusquality[]   = {4000,  68600, 133200, 197800, 262400, 327000, 391600, 456200, 520800, 585400, 650000};
	gint   flacquality[]   = {0,     1,     2,      3,      4,      5,      6,      7,      8,      9,      9};
	gfloat mp3quality[]    = {9.999, 9.0,   8.0,    7.0,    6.0,    5.0,    4.0,    3.0,    2.0,    1.0,    0.0};
	gint   aacquality[]    = {1,     100,   200,    300,    400,    500,    600,    700,    800,    900,    1000};
	/*Encoders parameters mappings*/
	typedef struct {
		gchar *encodertype;
		gchar *muxertype;
		gpointer quality;
		gboolean intquality;
		gboolean tagsinmuxer;
	} GstEncoderParams;
	GstEncoderParams encparams[] = {
		{NULL,          NULL,        NULL,           FALSE,  FALSE},
		{"wavenc",      NULL,        NULL,           FALSE,  FALSE},
		{"vorbisenc",   "oggmux",    vorbisquality,  FALSE,  FALSE},
		{"opusenc",     "oggmux",    opusquality,    TRUE,   FALSE},
		{"flacenc",     NULL,        flacquality,    TRUE,   FALSE},
		{"lamemp3enc",  "id3v2mux",  mp3quality,     FALSE,  TRUE},
		{"faac",        "mp4mux",    aacquality,     TRUE,   FALSE}
	};
	
	g_return_if_fail(self != NULL);
	g_return_if_fail(srcpad != NULL);
	g_return_if_fail(self->target != LARK_GSTREAMER_TARGET_AUDIO);
	
	/*Unlinking old encoding chain*/
	if (self->encoder != NULL) {
		sinkpad = gst_element_get_static_pad(self->encoder, "sink");
		gst_pad_unlink(srcpad, sinkpad);
		gst_object_unref(sinkpad);
	} else {
		if (self->muxer != NULL) {
			sinkpad = gst_element_get_static_pad(self->muxer, "sink");
			gst_pad_unlink(srcpad, sinkpad);
			gst_object_unref(sinkpad);
		} else if (self->filesink != NULL) {
			sinkpad = gst_element_get_static_pad(self->filesink, "sink");
			gst_pad_unlink(srcpad, sinkpad);
			gst_object_unref(sinkpad);
		}
	}
	/*Save pointers to old elements*/
	if (self->encoder != NULL) {
		oldencoder = self->encoder;
	} else {
		oldencoder = NULL;
	}
	if (self->muxer != NULL) {
		oldmuxer = self->muxer;
	} else {
		oldmuxer = NULL;
	}
	if (self->filesink != NULL) {
		oldfilesink = self->filesink;
	} else {
		oldfilesink = NULL;
	}
	/*Create new elements*/
	/*Encoder*/
	if (encparams[self->target].encodertype != NULL) {
		elementname = lark_gstreamer_build_element_name(self, "encoder");
		self->encoder = gst_element_factory_make(encparams[self->target].encodertype, elementname);
		/*Set quality*/
		if (encparams[self->target].quality != NULL) {
			if (encparams[self->target].intquality) {
				iquality = (gint *)(encparams[self->target].quality);
				g_object_set(self->encoder, "quality", iquality[self->quality], NULL);
			} else {
				fquality = (gfloat *)(encparams[self->target].quality);
				g_object_set(self->encoder, "quality", fquality[self->quality], NULL);
			}
		}
		/*Append tags in encoder*/
		if (!encparams[self->target].tagsinmuxer) {
			if (self->taglist != NULL) {
				gst_tag_setter_merge_tags(GST_TAG_SETTER(self->encoder), self->taglist, GST_TAG_MERGE_APPEND);
			}
		}
		g_free(elementname);
	} else {
		self->encoder = NULL;
	}
	/*Muxer*/
	if (encparams[self->target].muxertype != NULL) {
		elementname = lark_gstreamer_build_element_name(self, "muxer");
		self->muxer = gst_element_factory_make(encparams[self->target].muxertype, elementname);
		/*Append tags in muxer*/
		if (encparams[self->target].tagsinmuxer) {
			if (self->taglist != NULL) {
				gst_tag_setter_merge_tags(GST_TAG_SETTER(self->muxer), self->taglist, GST_TAG_MERGE_APPEND);
			}
		}
		g_free(elementname);
	} else {
		self->muxer = NULL;
	}
	/*File sink*/
	elementname = lark_gstreamer_build_element_name(self, "filesink");
	self->filesink = gst_element_factory_make("filesink", elementname);
	g_free(elementname);
	filepath = lark_gstreamer_build_file_path(self);
	g_object_set(self->filesink, "location", filepath, NULL);
	g_free(filepath);
	/*Callback for progress accounting*/
	sinkpad = gst_element_get_static_pad(self->filesink, "sink");
	gst_pad_add_probe(sinkpad, GST_PAD_PROBE_TYPE_BUFFER, (GstPadProbeCallback)lark_gstreamer_inspect_output_buffer, self, NULL);
	gst_object_unref(sinkpad);
	/*Pipeline construction*/
	if (self->encoder != NULL) {
		gst_bin_add(GST_BIN(self->pipeline), self->encoder);
	}
	if (self->muxer != NULL) {
		gst_bin_add(GST_BIN(self->pipeline), self->muxer);
	}
	gst_bin_add(GST_BIN(self->pipeline), self->filesink);
	/*State change*/
	if (self->encoder != NULL) {
		gst_element_sync_state_with_parent(self->encoder);
	}
	if (self->muxer != NULL) {
		gst_element_sync_state_with_parent(self->muxer);
	}
	gst_element_sync_state_with_parent(self->filesink);
	/*Linkage*/
	if (self->encoder != NULL) {
		if (self->muxer != NULL) {
			gst_element_link_many(self->encoder, self->muxer, self->filesink, NULL);
		} else {
			gst_element_link_many(self->encoder, self->filesink, NULL);
		} 
		sinkpad = gst_element_get_static_pad(self->encoder, "sink");
	} else {
		if (self->muxer != NULL) {
			gst_element_link_many(self->muxer, self->filesink, NULL);
			sinkpad = gst_element_get_static_pad(self->muxer, "sink");
		} else {
			sinkpad = gst_element_get_static_pad(self->filesink, "sink");
		}
	}
	gst_pad_link(srcpad, sinkpad);
	gst_object_unref(sinkpad);
	/*Old elements state update*/
	if (oldencoder != NULL) {
		gst_element_set_state(oldencoder, GST_STATE_NULL);
	}
	if (oldmuxer != NULL) {
		gst_element_set_state(oldmuxer, GST_STATE_NULL);
	}
	if (oldfilesink != NULL) {
		gst_element_set_state(oldfilesink, GST_STATE_NULL);
	}
	/*Old elements cleanup*/
	if (oldencoder != NULL) {
		gst_bin_remove(GST_BIN(self->pipeline), oldencoder);
	}
	if (oldmuxer != NULL) {
		gst_bin_remove(GST_BIN(self->pipeline), oldmuxer);
	}
	if (oldfilesink != NULL) {
		gst_bin_remove(GST_BIN(self->pipeline), oldfilesink);
	}
}

static GstPadProbeReturn lark_gstreamer_switch_output_file(GstPad *pad, GstPadProbeInfo *info, LarkGstreamer *self)
{
	g_return_val_if_fail(self != NULL, GST_PAD_PROBE_REMOVE);
	g_return_val_if_fail(pad  != NULL, GST_PAD_PROBE_REMOVE);
	g_return_val_if_fail(info != NULL, GST_PAD_PROBE_REMOVE);
	
	/*Rebuild pipeline tail*/
	lark_gstreamer_build_encoding_chain(self, pad);
	
	return GST_PAD_PROBE_REMOVE;
}

static GstPadProbeReturn lark_gstreamer_inspect_output_buffer(GstPad *pad, GstPadProbeInfo *info, LarkGstreamer *self)
{
	GstBuffer *buffer;
	GstPad *encpad;
	
	g_return_val_if_fail(self != NULL, GST_PAD_PROBE_OK);
	g_return_val_if_fail(info != NULL, GST_PAD_PROBE_OK);
	
	buffer = GST_PAD_PROBE_INFO_BUFFER(info);
	
	if (buffer != NULL) {
		/*Update local counters*/
		self->filesize += gst_buffer_get_size(buffer);
		if (GST_BUFFER_DURATION_IS_VALID(buffer)) {
			self->fileduration += GST_BUFFER_DURATION(buffer);
		}
		/*Handle limitations*/
		if ((self->target != LARK_GSTREAMER_TARGET_AUDIO) && (self->limitmode != LARK_GSTREAMER_LIMIT_MODE_NONE)) {
			if (((self->limitmode == LARK_GSTREAMER_LIMIT_MODE_SIZE) && (self->filesize >= self->sizelimit)) || ((self->limitmode == LARK_GSTREAMER_LIMIT_MODE_TIME) && (self->fileduration / GST_SECOND >= self->timelimit))) {
				/*Update global counters*/
				self->fullsize += self->filesize;
				self->filesize = 0;
				self->fullduration += self->fileduration;
				self->fileduration = 0;
				self->numfragments++;
				/*Block encoder's output pad and start new file in callback*/
				encpad = gst_element_get_static_pad(self->equalizer, "src");
				gst_pad_add_probe(encpad, GST_PAD_PROBE_TYPE_BLOCK_DOWNSTREAM, (GstPadProbeCallback)lark_gstreamer_switch_output_file, self, NULL);
				gst_object_unref(encpad);
			}
		}
	}
	
	return GST_PAD_PROBE_OK;
}

static void lark_gstreamer_message_handler(GstBus *bus, GstMessage *message, LarkGstreamer *self)
{
	GError *error;
	gchar *debug;
	
	g_return_if_fail(self != NULL);
	
	error = NULL;
	
	g_debug("GStreamer message: %s", gst_message_type_get_name(message->type));
	
	switch (GST_MESSAGE_TYPE(message)) {
		case GST_MESSAGE_LATENCY:
			gst_bin_recalculate_latency(GST_BIN(self->pipeline));
			break;
		case GST_MESSAGE_ERROR:
			/*Error*/
			gst_message_parse_error(message, &error, &debug);
			g_debug("GStreamer: %s", debug);
			g_free(debug);
			gst_tag_list_unref(self->taglist);
			g_signal_emit_by_name(G_OBJECT(self), "finished", FALSE, error);
			break;
		case GST_MESSAGE_EOS:
			g_debug("GStreamer EOS");
			gst_tag_list_unref(self->taglist);
			/*End of stream*/
			g_signal_emit_by_name(G_OBJECT(self), "finished", TRUE, NULL);
			/*Free elements*/
			gst_element_set_state(GST_ELEMENT(self->pipeline), GST_STATE_NULL);
			break;
		default:
			/*Unhandled message*/
			break;
	}
}

gboolean lark_gstreamer_start(LarkGstreamer *self)
{
	GDate *date;
	GstAudioInfo info;
	GstCaps *audio_caps;
	GstStateChangeReturn ret;
	GstBus *bus;
	GstPad *srcpad, *sinkpad;
	gint audioformat, i;
	/*Equalizer*/
	GObject *band;
	typedef struct {
		gfloat freq;
		gfloat width;
		gfloat *gain;
	} GstEqualizerBandState;
	GstEqualizerBandState eqstate[] = {
		{ 100.0,    50.0,  &self->eq100gain},
		{ 300.0,   100.0,  &self->eq300gain},
		{1000.0,   500.0,  &self->eq1kgain},
		{3000.0,  1000.0,  &self->eq3kgain},
		{8000.0,  2000.0,  &self->eq8kgain}
	};
			
	g_return_val_if_fail(self != NULL, FALSE);
	
	/*Zero counters*/
	self->numsamples = 0;
	self->numfragments = 0;
	self->fullsize = 0;
	self->filesize = 0;
	self->fullduration = 0;
	self->fileduration = 0;
	
	/*Metatags*/
	self->taglist = gst_tag_list_new_empty();
	if (self->titletag != NULL) {
		gst_tag_list_add(self->taglist, GST_TAG_MERGE_KEEP, GST_TAG_TITLE, self->titletag, NULL);
	}
	if (self->authortag != NULL) {
		gst_tag_list_add(self->taglist, GST_TAG_MERGE_KEEP, GST_TAG_ARTIST, self->authortag, NULL);
	}
	if (self->commenttag != NULL) {
		gst_tag_list_add(self->taglist, GST_TAG_MERGE_KEEP, GST_TAG_COMMENT, self->commenttag, NULL);
	}
	date = g_date_new();
	g_date_set_time_t(date, time(NULL));
	gst_tag_list_add(self->taglist, GST_TAG_MERGE_KEEP, GST_TAG_DATE, date, NULL);
	g_date_free (date);
	
	/*Pipeline*/
	self->pipeline = GST_PIPELINE(gst_pipeline_new("pipeline"));
	
	/*Elements*/
	self->appsrc = gst_element_factory_make("appsrc", "source");
	self->converter = gst_element_factory_make("audioconvert", "converter");
	self->equalizer = gst_element_factory_make("equalizer-nbands", "equalizer");
	
	/*Equalizer*/
	g_object_set(G_OBJECT(self->equalizer), "num-bands", 5, NULL);
	for (i=LARK_GSTREAMER_EQ_FREQ_100; i<LARK_GSTREAMER_EQ_FREQ_COUNT; i++) {
		band = gst_child_proxy_get_child_by_index(GST_CHILD_PROXY(self->equalizer), i);
		g_object_set(G_OBJECT(band), "freq", eqstate[i].freq, "bandwidth", eqstate[i].width, "gain", *eqstate[i].gain, NULL);
		g_object_unref(G_OBJECT(band));
	}
	
	/*Encoder chain elements*/
	self->encoder = NULL;
	self->muxer = NULL;
	self->filesink = NULL;
	
	/*Input chain of pipeline*/
	gst_bin_add_many(GST_BIN(self->pipeline), self->appsrc, self->converter, self->equalizer, NULL);
	gst_element_link_many(self->appsrc, self->converter, self->equalizer, NULL);
	srcpad = gst_element_get_static_pad(self->equalizer, "src");
	
	/*Link elements*/
	if (self->target == LARK_GSTREAMER_TARGET_AUDIO) {
		/*Send output to audio system*/
		self->audiosink = gst_element_factory_make("pulsesink", "audiosink");
		/*Add callback for progress accounting*/
		sinkpad = gst_element_get_static_pad(self->audiosink, "sink");
		gst_pad_add_probe(sinkpad, GST_PAD_PROBE_TYPE_BUFFER, (GstPadProbeCallback)lark_gstreamer_inspect_output_buffer, self, NULL);
		/*Add/Link elements*/
		gst_bin_add(GST_BIN(self->pipeline), self->audiosink);
		gst_pad_link(srcpad, sinkpad);
		gst_object_unref(sinkpad);
	} else {
		/*Send output to file(s)*/
		lark_gstreamer_build_encoding_chain(self, srcpad);
	}
	
	gst_object_unref(srcpad);
	
	/*AppSource setup*/
	switch (self->format) {
		case LARK_GSTREAMER_SRC_FORMAT_U8:
			audioformat = GST_AUDIO_FORMAT_U8;
			break;
		case LARK_GSTREAMER_SRC_FORMAT_S8:
			audioformat = GST_AUDIO_FORMAT_S8;
			break;
		case LARK_GSTREAMER_SRC_FORMAT_U16LE:
			audioformat = GST_AUDIO_FORMAT_U16LE;
			break;
		case LARK_GSTREAMER_SRC_FORMAT_U16BE:
			audioformat = GST_AUDIO_FORMAT_U16BE;
			break;
		case LARK_GSTREAMER_SRC_FORMAT_S16LE:
			audioformat = GST_AUDIO_FORMAT_S16LE;
			break;
		case LARK_GSTREAMER_SRC_FORMAT_S16BE:
			audioformat = GST_AUDIO_FORMAT_S16BE;
			break;
		case LARK_GSTREAMER_SRC_FORMAT_U32LE:
			audioformat = GST_AUDIO_FORMAT_U16LE;
			break;
		case LARK_GSTREAMER_SRC_FORMAT_U32BE:
			audioformat = GST_AUDIO_FORMAT_U32BE;
			break;
		case LARK_GSTREAMER_SRC_FORMAT_S32LE:
			audioformat = GST_AUDIO_FORMAT_S32LE;
			break;
		case LARK_GSTREAMER_SRC_FORMAT_S32BE:
			audioformat = GST_AUDIO_FORMAT_S32BE;
			break;
		default:
			audioformat = GST_AUDIO_FORMAT_U8;
			break;
	}
	
	/*Audio caps*/
	gst_audio_info_set_format(&info, audioformat, self->samplerate, self->channels, NULL);
	audio_caps = gst_audio_info_to_caps(&info);
	g_object_set(self->appsrc, "block", TRUE, "max-bytes", (guint64)self->queuebytes, "caps", audio_caps, "format", GST_FORMAT_TIME, NULL);
	
	/*Attach signal handlers to the bus*/
	bus = gst_pipeline_get_bus(GST_PIPELINE(self->pipeline));
	gst_bus_add_signal_watch(bus);
	g_signal_connect(G_OBJECT(bus), "message", (GCallback)lark_gstreamer_message_handler, self);
	gst_object_unref(bus);
	
	/*Start GStreamer*/
	ret = gst_element_set_state(GST_ELEMENT(self->pipeline), GST_STATE_PLAYING);
	if (ret == GST_STATE_CHANGE_FAILURE) {
		gst_tag_list_unref(self->taglist);
		gst_object_unref(self->pipeline);
	}
	
	g_debug("GStreamer started\n");
	
	return TRUE;
}

LarkGstreamer *lark_gstreamer_new(GError **error)
{
	return g_initable_new(lark_gstreamer_get_type(), NULL, error, NULL);
}
