/*
 *      lark-pref-dialog.c
 *
 *      Copyright 2016-2018 Alex <alex@linuxonly.ru>
 *
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>

#include "lark-pref-dialog.h"

enum {
	LARK_PREF_VOICES_TREE_VIEW_SELECTED = 0,
	LARK_PREF_VOICES_TREE_VIEW_DESCRIPTION,
	LARK_PREF_VOICES_TREE_VIEW_PROFILE,
	LARK_PREF_VOICES_TREE_VIEW_COLUMNS
};

enum {
	PROP_0,
	PROP_CORE,
	PROP_RHVOICE,
	PROP_ESPEAK,
	PROP_PREFERENCES_SETTINGS,
	LAST_PROP
};

static GParamSpec *properties [LAST_PROP];

struct _LarkPrefDialog
{
	GtkDialog parent_instance;
	/*Objects*/
	LarkCore *core;
	LarkTtsEngine *rhvoice;
	LarkTtsEngine *espeak;
	GSettings *prefsettings;
	/*Widgets*/
	GtkTreeView *LarkPrefVoicesTreeView;
	GtkAdjustment *LarkPrefDialogPitchAdjustment;
	GtkAdjustment *LarkPrefDialogRateAdjustment;
	GtkAdjustment *LarkPrefDialogVolumeAdjustment;
	GtkComboBoxText *LarkPrefDialogCapitals;
	GtkComboBoxText *LarkPrefDialogPunctuation;
};

struct _LarkPrefDialogClass
{
	GtkDialogClass parent_class;
};

G_DEFINE_TYPE (LarkPrefDialog, lark_pref_dialog, GTK_TYPE_DIALOG)


static void lark_pref_dialog_get_property(GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
	LarkPrefDialog *self = (LarkPrefDialog *)object;
	
	switch (prop_id) {
		case PROP_CORE:
			g_value_set_pointer(value, self->core);
			break;
		case PROP_RHVOICE:
			g_value_set_pointer(value, self->rhvoice);
			break;
		case PROP_ESPEAK:
			g_value_set_pointer(value, self->espeak);
			break;
		case PROP_PREFERENCES_SETTINGS:
			g_value_set_pointer(value, self->prefsettings);
			break;
	}
}

static void lark_pref_dialog_set_property(GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
	LarkPrefDialog *self = (LarkPrefDialog *)object;
	
	switch (prop_id) {
		case PROP_CORE:
			self->core = g_value_get_pointer(value);
			break;
		case PROP_RHVOICE:
			self->rhvoice = g_value_get_pointer(value);
			break;
		case PROP_ESPEAK:
			self->espeak = g_value_get_pointer(value);
			break;
		case PROP_PREFERENCES_SETTINGS:
			self->prefsettings = g_value_get_pointer(value);
			break;
	}
}

static void lark_pref_dialog_pitch_value_changed(GtkAdjustment *adjustment, LarkPrefDialog *self)
{
	gdouble pitch;
	
	pitch = gtk_adjustment_get_value(adjustment);
	
	lark_tts_engine_set_pitch(LARK_TTS_ENGINE(self->rhvoice), pitch);
	lark_tts_engine_set_pitch(LARK_TTS_ENGINE(self->espeak), pitch);
	
	g_settings_set_double(self->prefsettings, "pitch", pitch);
}

static void lark_pref_dialog_rate_value_changed(GtkAdjustment *adjustment, LarkPrefDialog *self)
{
	gdouble rate;
	
	rate = gtk_adjustment_get_value(adjustment);
	
	lark_tts_engine_set_pitch(LARK_TTS_ENGINE(self->rhvoice), rate);
	lark_tts_engine_set_pitch(LARK_TTS_ENGINE(self->espeak), rate);
	
	g_settings_set_double(self->prefsettings, "rate", rate);
}

static void lark_pref_dialog_volume_value_changed(GtkAdjustment *adjustment, LarkPrefDialog *self)
{
	gdouble volume;
	
	volume = gtk_adjustment_get_value(adjustment);
	
	lark_tts_engine_set_pitch(LARK_TTS_ENGINE(self->rhvoice), volume);
	lark_tts_engine_set_pitch(LARK_TTS_ENGINE(self->espeak), volume);
	
	g_settings_set_double(self->prefsettings, "volume", volume);
}

static void lark_pref_dialog_capitals_indication_changed(GtkComboBox *widget, LarkPrefDialog *self)
{
	gint indication;
	gchar *strparam;
	
	indication = gtk_combo_box_get_active(widget);
	
	switch (indication) {
		case LARK_TTS_CAPITALS_SYMBOLS_INDICATION_DEFAULT:
			strparam = "default";
			break;
		case LARK_TTS_CAPITALS_SYMBOLS_INDICATION_PITCH:
			strparam = "pitch";
			break;
		case LARK_TTS_CAPITALS_SYMBOLS_INDICATION_VOLUME:
			strparam = "volume";
			break;
		case LARK_TTS_CAPITALS_SYMBOLS_INDICATION_OFF:
			strparam = "off";
			break;
		default:
			strparam = NULL;
			break;
	}
	
	if (strparam != NULL) {
		lark_tts_engine_set_capitals_indication(LARK_TTS_ENGINE(self->rhvoice), indication);
		lark_tts_engine_set_capitals_indication(LARK_TTS_ENGINE(self->espeak), indication);
		g_settings_set_string(self->prefsettings, "capitals-symbols-indication", strparam);
	}
}
               
static void lark_pref_dialog_punctuation_indication_changed(GtkComboBox *widget, LarkPrefDialog *self)
{
	gint indication;
	gchar *strparam;
	
	indication = gtk_combo_box_get_active(widget);
	
	switch (indication) {
		case LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_DEFAULT:
			strparam = "default";
			break;
		case LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_ON:
			strparam = "on";
			break;
		case LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_OFF:
			strparam = "off";
			break;
		default:
			strparam = NULL;
			break;
	}
	
	if (strparam != NULL) {
		lark_tts_engine_set_punctuation_indication(LARK_TTS_ENGINE(self->rhvoice), indication);
		lark_tts_engine_set_punctuation_indication(LARK_TTS_ENGINE(self->espeak), indication);
		g_settings_set_string(self->prefsettings, "punctuation-symbols-indication", strparam);
	}
}

static void lark_pref_dialog_constructed(GObject *object)
{
	GtkTreeModel *model;
	GtkTreeIter iter;
	gchar *descr, *profile, *selstring;
	GSList *voices[2], *voicenode;
	gboolean profileselected, profilefound;
	gint e;
	gchar *profilespecs[2] = {"rhvoice::%s", "espeak::%s"};
	LarkTtsEngineVoice *voice;
	
	LarkPrefDialog *self = (LarkPrefDialog *)object;
	
	selstring = g_settings_get_string(self->prefsettings, "voice-profile");
	
	model = gtk_tree_view_get_model(self->LarkPrefVoicesTreeView);
	
	if (model != NULL) {
		profilefound = FALSE;
		voices[0] = lark_tts_engine_get_voices(self->rhvoice);
		voices[1] = lark_tts_engine_get_voices(self->espeak);
		for (e = 0; e < 2; e++) {
			for (voicenode = voices[e]; voicenode->next != NULL; voicenode = voicenode->next) {
				voice = (LarkTtsEngineVoice *)voicenode->data;
				/*Description*/
				descr = lark_tts_engine_format_voice_description(voice);
				/*Profile*/
				profile = g_strdup_printf(profilespecs[e], voice->name);
				profileselected = FALSE;
				if (!profilefound) {
					if (g_strcmp0(profile, selstring) == 0) {
						profileselected = TRUE;
						profilefound = TRUE;
					}
				}
				/*Add to list*/			
				gtk_list_store_append(GTK_LIST_STORE(model), &iter);
				gtk_list_store_set(GTK_LIST_STORE(model), &iter, LARK_PREF_VOICES_TREE_VIEW_SELECTED, profileselected, LARK_PREF_VOICES_TREE_VIEW_DESCRIPTION, descr, LARK_PREF_VOICES_TREE_VIEW_PROFILE, profile, -1);
				/*Free resources*/
				g_free(profile);
				g_free(descr);
			}
		}
	}
	g_free(selstring);
	
	/*Set synthesis parameters*/
	gtk_adjustment_set_value(self->LarkPrefDialogPitchAdjustment, g_settings_get_double(self->prefsettings, "pitch"));
	gtk_adjustment_set_value(self->LarkPrefDialogRateAdjustment, g_settings_get_double(self->prefsettings, "rate"));
	gtk_adjustment_set_value(self->LarkPrefDialogVolumeAdjustment, g_settings_get_double(self->prefsettings, "volume"));
	
	selstring = g_settings_get_string(self->prefsettings, "capitals-symbols-indication");
	if (g_strcmp0(selstring, "pitch") == 0) {
		gtk_combo_box_set_active(GTK_COMBO_BOX(self->LarkPrefDialogCapitals), LARK_TTS_CAPITALS_SYMBOLS_INDICATION_PITCH);
	} else if (g_strcmp0(selstring, "volume") == 0) {
		gtk_combo_box_set_active(GTK_COMBO_BOX(self->LarkPrefDialogCapitals), LARK_TTS_CAPITALS_SYMBOLS_INDICATION_VOLUME);
	} else if (g_strcmp0(selstring, "off") == 0) {
		gtk_combo_box_set_active(GTK_COMBO_BOX(self->LarkPrefDialogCapitals), LARK_TTS_CAPITALS_SYMBOLS_INDICATION_OFF);
	} else {
		gtk_combo_box_set_active(GTK_COMBO_BOX(self->LarkPrefDialogCapitals), LARK_TTS_CAPITALS_SYMBOLS_INDICATION_DEFAULT);
	}
	g_free(selstring);
	
	selstring = g_settings_get_string(self->prefsettings, "punctuation-symbols-indication");
	if (g_strcmp0(selstring, "on") == 0) {
		gtk_combo_box_set_active(GTK_COMBO_BOX(self->LarkPrefDialogPunctuation), LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_ON);
	} else if (g_strcmp0(selstring, "off") == 0) {
		gtk_combo_box_set_active(GTK_COMBO_BOX(self->LarkPrefDialogPunctuation), LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_OFF);
	} else {
		gtk_combo_box_set_active(GTK_COMBO_BOX(self->LarkPrefDialogPunctuation), LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_DEFAULT);
	}
	g_free(selstring);
}

static void lark_pref_dialog_class_init(LarkPrefDialogClass *klass)
{
	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS(klass);
	GObjectClass *object_class = G_OBJECT_CLASS(klass);
	
	object_class->get_property = lark_pref_dialog_get_property;
	object_class->set_property = lark_pref_dialog_set_property;
	object_class->constructed = lark_pref_dialog_constructed;
	
	properties[PROP_CORE] =
		g_param_spec_pointer("core",
							"Core object",
							"Pointer to core object",
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
	
	properties[PROP_RHVOICE] =
		g_param_spec_pointer("rhvoice",
							"RHVoice object",
							"Pointer to RHVoice object",
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
							
	properties[PROP_ESPEAK] =
		g_param_spec_pointer("espeak",
							"Espeak object",
							"Pointer to espeak object",
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
							
	properties[PROP_PREFERENCES_SETTINGS] =
		g_param_spec_pointer("preferences-settings",
							"Preferences settings object",
							"Pointer to preferences settings object",
							G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE);
								
	g_object_class_install_properties(object_class, LAST_PROP, properties);
	
	gtk_widget_class_set_template_from_resource(widget_class, "/ru/linuxonly/lark/ui/pref-dialog.ui");
	
	gtk_widget_class_bind_template_child(widget_class, LarkPrefDialog, LarkPrefVoicesTreeView);
	
	gtk_widget_class_bind_template_child(widget_class, LarkPrefDialog, LarkPrefDialogPitchAdjustment);
	gtk_widget_class_bind_template_child(widget_class, LarkPrefDialog, LarkPrefDialogRateAdjustment);
	gtk_widget_class_bind_template_child(widget_class, LarkPrefDialog, LarkPrefDialogVolumeAdjustment);
	gtk_widget_class_bind_template_child(widget_class, LarkPrefDialog, LarkPrefDialogCapitals);
	gtk_widget_class_bind_template_child(widget_class, LarkPrefDialog, LarkPrefDialogPunctuation);
		
	gtk_widget_class_bind_template_callback(widget_class, lark_pref_dialog_pitch_value_changed);
	gtk_widget_class_bind_template_callback(widget_class, lark_pref_dialog_rate_value_changed);
	gtk_widget_class_bind_template_callback(widget_class, lark_pref_dialog_volume_value_changed);
	gtk_widget_class_bind_template_callback(widget_class, lark_pref_dialog_capitals_indication_changed);
	gtk_widget_class_bind_template_callback(widget_class, lark_pref_dialog_punctuation_indication_changed);
}

static void lark_pref_dialog_select_voice(GtkCellRendererToggle *cell_renderer, gchar *path, LarkPrefDialog *self)
{
	GtkTreeModel *model;
	GtkTreeIter iter;
	GtkTreePath *curpath, *selpath;
	gchar *profile;
	
	g_return_if_fail(self != NULL);
	
	model = gtk_tree_view_get_model(self->LarkPrefVoicesTreeView);
	
	if (model != NULL) {
		selpath = gtk_tree_path_new_from_string((const gchar *)path);
		if (selpath != NULL) {
			if (gtk_tree_model_get_iter_first(model, &iter)) {
				do {
					curpath = gtk_tree_model_get_path(model, &iter);
					if (curpath != NULL) {
						if (gtk_tree_path_compare((const GtkTreePath *)curpath, (const GtkTreePath *)selpath) == 0) {
							/*Save selected profile*/
							gtk_tree_model_get(model, &iter, LARK_PREF_VOICES_TREE_VIEW_PROFILE, &profile, -1);
							g_settings_set_string(self->prefsettings, "voice-profile", profile);
							g_free(profile);
							/*Select profile*/
							gtk_list_store_set(GTK_LIST_STORE(model), &iter, LARK_PREF_VOICES_TREE_VIEW_SELECTED, TRUE, -1);
						} else {
							gtk_list_store_set(GTK_LIST_STORE(model), &iter, LARK_PREF_VOICES_TREE_VIEW_SELECTED, FALSE, -1);
						}
						gtk_tree_path_free(curpath);
					}
				} while (gtk_tree_model_iter_next(model, &iter));
			}
			gtk_tree_path_free(selpath);
		}
	}
}

static void lark_pref_dialog_init(LarkPrefDialog *self)
{
	GtkCellRenderer *renderer;
	GtkTreeViewColumn *column;
	
	gtk_widget_init_template(GTK_WIDGET(self));
	
	renderer = gtk_cell_renderer_toggle_new();
	gtk_cell_renderer_toggle_set_radio(GTK_CELL_RENDERER_TOGGLE(renderer), TRUE);
	g_signal_connect(G_OBJECT(renderer), "toggled", (GCallback)lark_pref_dialog_select_voice, self);
	column = gtk_tree_view_column_new_with_attributes("Selected", renderer, "active", LARK_PREF_VOICES_TREE_VIEW_SELECTED, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(self->LarkPrefVoicesTreeView), column);
	
	renderer = gtk_cell_renderer_text_new();
	column = gtk_tree_view_column_new_with_attributes("Description", renderer, "markup", LARK_PREF_VOICES_TREE_VIEW_DESCRIPTION, NULL);
	gtk_tree_view_append_column(GTK_TREE_VIEW(self->LarkPrefVoicesTreeView), column);
	
}

GtkWidget *lark_pref_dialog_new(GtkWindow *parent, LarkCore *core, LarkTtsEngine *rhvoice, LarkTtsEngine *espeak, GSettings *prefsettings)
{
	return g_object_new(lark_pref_dialog_get_type(), "transient-for", parent, "use-header-bar", TRUE, "core", core, "rhvoice", rhvoice, "espeak", espeak, "preferences-settings", prefsettings, NULL);
}
