/*
 *      lark-tts-engine.c
 *
 *      Copyright 2016-2018 Alex <alex@linuxonly.ru>
 *
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdio.h>
#include <stdlib.h>

#include "lark-tts-engine.h"

typedef struct _LarkTtsEnginePrivate {
	GObject parent_instance;
	gint format;
	gint samplerate;
	gint channels;
	GSList *voices;
	gchar *voiceprofile;
	gdouble rate;
	gdouble pitch;
	gdouble volume;
	gint capitalsind;
	gint punctuationind;
} LarkTtsEnginePrivate;

G_DEFINE_TYPE_WITH_PRIVATE (LarkTtsEngine, lark_tts_engine, G_TYPE_OBJECT)

enum {
	PROP_0,
	PROP_FORMAT,
	PROP_SAMPLERATE,
	PROP_CHANNELS,
	PROP_VOICES,
	PROP_VOICE_PROFILE,
	PROP_RATE,
	PROP_PITCH,
	PROP_VOLUME,
	PROP_CAPITALS_IND,
	PROP_PUNCTUATION_IND,
	LAST_PROP
};

static GParamSpec *properties [LAST_PROP];

enum {
	SIGNAL_SAMPLES_GENERATED,
	LAST_SIGNAL
};

static guint signals [LAST_SIGNAL];

static void lark_tts_engine_get_property (GObject *object, guint prop_id, GValue *value, GParamSpec *pspec)
{
	LarkTtsEngine *self = (LarkTtsEngine *)object;
	LarkTtsEnginePrivate *priv;
	
	priv = lark_tts_engine_get_instance_private (self);
	
	switch (prop_id) {
		case PROP_VOICES:
			g_value_set_pointer (value, priv->voices);
			break;
		case PROP_FORMAT:
			g_value_set_int (value, priv->format);
			break;
		case PROP_SAMPLERATE:
			g_value_set_int (value, priv->samplerate);
			break;
		case PROP_CHANNELS:
			g_value_set_int (value, priv->channels);
			break;
		case PROP_VOICE_PROFILE:
			g_value_set_string(value, priv->voiceprofile);
			break;
		case PROP_RATE:
			g_value_set_double(value, priv->rate);
			break;
		case PROP_PITCH:
			g_value_set_double(value, priv->pitch);
			break;
		case PROP_VOLUME:
			g_value_set_double(value, priv->volume);
			break;
		case PROP_CAPITALS_IND:
			g_value_set_int (value, priv->capitalsind);
			break;
		case PROP_PUNCTUATION_IND:
			g_value_set_int (value, priv->punctuationind);
			break;
		default:
			break;
	}
}

static void lark_tts_engine_set_property (GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec)
{
	LarkTtsEngine *self = (LarkTtsEngine *)object;
	LarkTtsEnginePrivate *priv;
	
	priv = lark_tts_engine_get_instance_private (self);
	
	switch (prop_id) {
		case PROP_VOICES:
			priv->voices = g_value_get_pointer (value);
			break;
		case PROP_FORMAT:
			priv->format = g_value_get_int (value);
			break;
		case PROP_SAMPLERATE:
			priv->samplerate = g_value_get_int (value);
			break;
		case PROP_CHANNELS:
			priv->channels = g_value_get_int (value);
			break;
		case PROP_VOICE_PROFILE:
			if (priv->voiceprofile != NULL) {
				g_free(priv->voiceprofile);
			}
			priv->voiceprofile = g_strdup(g_value_get_string(value));
			break;
		case PROP_RATE:
			priv->rate = g_value_get_double (value);
			break;
		case PROP_PITCH:
			priv->pitch = g_value_get_double (value);
			break;
		case PROP_VOLUME:
			priv->volume = g_value_get_double (value);
			break;
		case PROP_CAPITALS_IND:
			priv->capitalsind = g_value_get_int (value);
			break;
		case PROP_PUNCTUATION_IND:
			priv->punctuationind = g_value_get_int (value);
			break;
		default:
			break;
	}
}

static void lark_tts_engine_finalize (GObject *object)
{
	LarkTtsEngine *self = (LarkTtsEngine *)object;
	LarkTtsEnginePrivate *priv;
	GSList *voicenode;
	LarkTtsEngineVoice *voice;
	
	priv = lark_tts_engine_get_instance_private (self);
	
	if (priv->voices != NULL) {
		for (voicenode = priv->voices; voicenode->next != NULL; voicenode = voicenode->next) {
			voice = (LarkTtsEngineVoice *)voicenode->data;
			g_free(voice->name);
			g_ptr_array_free(voice->languages, TRUE);
			g_array_free(voice->genders, TRUE);
			g_free(voice);
		}
		g_slist_free(priv->voices);
	}
	
	G_OBJECT_CLASS (lark_tts_engine_parent_class)->finalize (object);
}

static void lark_tts_engine_class_init (LarkTtsEngineClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS(klass);
	
	object_class->get_property = lark_tts_engine_get_property;
	object_class->set_property = lark_tts_engine_set_property;
	object_class->finalize     = lark_tts_engine_finalize;
	
	properties[PROP_FORMAT] =
		g_param_spec_int("format",
							"Wave data format",
							"Wave data format",
							LARK_GSTREAMER_SRC_FORMAT_U8,
							LARK_GSTREAMER_SRC_FORMAT_S32BE,
							LARK_GSTREAMER_SRC_FORMAT_S16LE,
							G_PARAM_READWRITE);
							
	properties[PROP_SAMPLERATE] =
		g_param_spec_int("sample-rate",
							"Wave data sample rate",
							"Wave data sample rate",
							LARK_GSTREAMER_MIN_SAMPERATE,
							LARK_GSTREAMER_MAX_SAMPERATE,
							LARK_GSTREAMER_DEF_SAMPERATE,
							G_PARAM_READWRITE);
							
	properties[PROP_CHANNELS] =
		g_param_spec_int("channels",
							"Wave channels number",
							"Wave channels number",
							LARK_GSTREAMER_MIN_CHANNELS,
							LARK_GSTREAMER_MAX_CHANNELS,
							LARK_GSTREAMER_DEF_CHANNELS,
							G_PARAM_READWRITE);
	
	properties [PROP_VOICES] =
		g_param_spec_pointer ("voices",
							"Voices list",
							"Voices list",
							G_PARAM_READWRITE);
	
	properties [PROP_VOICE_PROFILE] =
		g_param_spec_string ("voice-profile",
							"Voice profile",
							"Voice profile",
							"default",
							G_PARAM_READWRITE);
	
	properties [PROP_RATE] =
		g_param_spec_double ("rate",
							"Speech rate",
							"Speech rate",
							-1.0,
							1.0,
							0.0,							
							G_PARAM_READWRITE);
	
	properties [PROP_PITCH] =
		g_param_spec_double ("pitch",
							"Speech pitch",
							"Speech pitch",
							-1.0,
							1.0,
							0.0,							
							G_PARAM_READWRITE);
							
	properties [PROP_VOLUME] =
		g_param_spec_double ("volume",
							"Speech volume",
							"Speech volume",
							-1.0,
							1.0,
							0.0,							
							G_PARAM_READWRITE);
	
	properties [PROP_CAPITALS_IND] =
		g_param_spec_int("capitals-indication",
							"Capitals indication format",
							"Capitals indication format",
							LARK_TTS_CAPITALS_SYMBOLS_INDICATION_DEFAULT,
							LARK_TTS_CAPITALS_SYMBOLS_INDICATION_OFF,
							LARK_TTS_CAPITALS_SYMBOLS_INDICATION_DEFAULT,
							G_PARAM_READWRITE);
	
	
	properties [PROP_PUNCTUATION_IND] =
		g_param_spec_int("punctuation-indication",
							"Punctuation indication format",
							"Punctuation indication format",
							LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_DEFAULT,
							LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_OFF,
							LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_DEFAULT,
							G_PARAM_READWRITE);
	
	
	signals[SIGNAL_SAMPLES_GENERATED] =
		g_signal_new("samples-generated",
							G_TYPE_FROM_CLASS(klass),
							G_SIGNAL_RUN_LAST,
							0,
							NULL,
							NULL,
							g_cclosure_marshal_VOID__UINT_POINTER,
							G_TYPE_NONE,
							2,
							G_TYPE_UINT,
							G_TYPE_POINTER);
	
	g_object_class_install_properties (object_class, LAST_PROP, properties);
}

static void lark_tts_engine_init (LarkTtsEngine *self)
{
	
}

LarkTtsEngine *lark_tts_engine_new (void)
{
	return g_object_new (lark_tts_engine_get_type (), NULL);
}

GSList *lark_tts_engine_get_voices(LarkTtsEngine *engine)
{
	LarkTtsEnginePrivate *priv;
	
	g_return_val_if_fail(engine != NULL, NULL);
	
	priv = lark_tts_engine_get_instance_private (engine);
	
	return priv->voices;
}

gint lark_tts_engine_get_format(LarkTtsEngine *engine)
{
	LarkTtsEnginePrivate *priv;
	
	g_return_val_if_fail(engine != NULL, LARK_GSTREAMER_SRC_FORMAT_S16LE);
	
	priv = lark_tts_engine_get_instance_private (engine);
	
	return priv->format;
}

gint lark_tts_engine_get_samplerate(LarkTtsEngine *engine)
{
	LarkTtsEnginePrivate *priv;
	
	g_return_val_if_fail(engine != NULL, LARK_GSTREAMER_DEF_SAMPERATE);
	
	priv = lark_tts_engine_get_instance_private (engine);
	
	return priv->samplerate;
}

gint lark_tts_engine_get_channels(LarkTtsEngine *engine)
{
	LarkTtsEnginePrivate *priv;
	
	g_return_val_if_fail(engine != NULL, LARK_GSTREAMER_DEF_CHANNELS);
	
	priv = lark_tts_engine_get_instance_private (engine);
	
	return priv->channels;
}

void lark_tts_engine_set_voice_profile(LarkTtsEngine *engine, gchar *voiceprofile)
{
	LarkTtsEnginePrivate *priv;
	
	g_return_if_fail(engine != NULL);
	g_return_if_fail(voiceprofile != NULL);
	
	priv = lark_tts_engine_get_instance_private (engine);
	
	if (priv->voiceprofile != NULL) {
		g_free(priv->voiceprofile);
	}
	
	priv->voiceprofile = g_strdup(voiceprofile);
	
	g_object_notify_by_pspec(G_OBJECT(engine), properties[PROP_VOICE_PROFILE]);
}

gchar *lark_tts_engine_get_voice_profile(LarkTtsEngine *engine)
{
	LarkTtsEnginePrivate *priv;
	
	g_return_val_if_fail(engine != NULL, "default");
	
	priv = lark_tts_engine_get_instance_private (engine);
	
	return priv->voiceprofile;
}

void lark_tts_engine_set_rate(LarkTtsEngine *engine, gdouble rate)
{
	LarkTtsEnginePrivate *priv;
	
	g_return_if_fail(engine != NULL);
	g_return_if_fail(rate >= -1.0 && rate <= 1.0);
	
	priv = lark_tts_engine_get_instance_private (engine);
	
	priv->rate = rate;
	
	g_object_notify_by_pspec(G_OBJECT(engine), properties[PROP_RATE]);
}

gdouble lark_tts_engine_get_rate(LarkTtsEngine *engine)
{
	LarkTtsEnginePrivate *priv;
	
	g_return_val_if_fail(engine != NULL, 0.0);
	
	priv = lark_tts_engine_get_instance_private (engine);
	
	return priv->rate;
}

void lark_tts_engine_set_pitch(LarkTtsEngine *engine, gdouble pitch)
{
	LarkTtsEnginePrivate *priv;
	
	g_return_if_fail(engine != NULL);
	g_return_if_fail(pitch >= -1.0 && pitch <= 1.0);
	
	priv = lark_tts_engine_get_instance_private (engine);
	
	priv->pitch = pitch;
	
	g_object_notify_by_pspec(G_OBJECT(engine), properties[PROP_PITCH]);
}

gdouble lark_tts_engine_get_pitch(LarkTtsEngine *engine)
{
	LarkTtsEnginePrivate *priv;
	
	g_return_val_if_fail(engine != NULL, 0.0);
	
	priv = lark_tts_engine_get_instance_private (engine);
	
	return priv->pitch;
}

void lark_tts_engine_set_volume(LarkTtsEngine *engine, gdouble volume)
{
	LarkTtsEnginePrivate *priv;
	
	g_return_if_fail(engine != NULL);
	g_return_if_fail(volume >= -1.0 && volume <= 1.0);
	
	priv = lark_tts_engine_get_instance_private (engine);
	
	priv->volume = volume;
	
	g_object_notify_by_pspec(G_OBJECT(engine), properties[PROP_VOLUME]);
}

gdouble lark_tts_engine_get_volume(LarkTtsEngine *engine)
{
	LarkTtsEnginePrivate *priv;
	
	g_return_val_if_fail(engine != NULL, 0.0);
	
	priv = lark_tts_engine_get_instance_private (engine);
	
	return priv->volume;
}

void lark_tts_engine_set_capitals_indication(LarkTtsEngine *engine, gint capitalsind)
{
	LarkTtsEnginePrivate *priv;
	
	g_return_if_fail(engine != NULL);
	g_return_if_fail(capitalsind >= LARK_TTS_CAPITALS_SYMBOLS_INDICATION_DEFAULT && capitalsind <= LARK_TTS_CAPITALS_SYMBOLS_INDICATION_OFF);
	
	priv = lark_tts_engine_get_instance_private (engine);
	
	priv->capitalsind = capitalsind;
	
	g_object_notify_by_pspec(G_OBJECT(engine), properties[PROP_CAPITALS_IND]);
}

gint lark_tts_engine_get_capitals_indication(LarkTtsEngine *engine)
{
	LarkTtsEnginePrivate *priv;
	
	g_return_val_if_fail(engine != NULL, LARK_TTS_CAPITALS_SYMBOLS_INDICATION_DEFAULT);
	
	priv = lark_tts_engine_get_instance_private (engine);
	
	return priv->capitalsind;
}

void lark_tts_engine_set_punctuation_indication(LarkTtsEngine *engine, gint punctuationind)
{
	LarkTtsEnginePrivate *priv;
	
	g_return_if_fail(engine != NULL);
	g_return_if_fail(punctuationind >= LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_DEFAULT && punctuationind <= LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_OFF);
	
	priv = lark_tts_engine_get_instance_private (engine);
	
	priv->punctuationind = punctuationind;
	
	g_object_notify_by_pspec(G_OBJECT(engine), properties[PROP_PUNCTUATION_IND]);
}

gint lark_tts_engine_get_punctuation_indication(LarkTtsEngine *engine)
{
	LarkTtsEnginePrivate *priv;
	
	g_return_val_if_fail(engine != NULL, LARK_TTS_PUNCTUATION_SYMBOLS_INDICATION_DEFAULT);
	
	priv = lark_tts_engine_get_instance_private (engine);
	
	return priv->punctuationind;
}

static gchar *lark_tts_engine_find_language_description(gchar *language)
{
	gint i;
	gchar *languages[139][2] = {
		{"aa",      "Afar"},
		{"af",      "Afrikaans"},
		{"am",      "Amharic"},
		{"an",      "Aragonese"},
		{"ar",      "Arabic"},
		{"as",      "Assamese"},
		{"ast",     "Asturian"},
		{"az",      "Azerbaijani"},
		{"be",      "Belarusian"},
		{"ber",     "Berber languages"},
		{"bg",      "Bulgarian"},
		{"bn",      "Bengali"},
		{"bo",      "Tibetan"},
		{"br",      "Breton"},
		{"bs",      "Bosnian"},
		{"ca",      "Catalan; Valencian"},
		{"crh",     "Turkish, Crimean"},
		{"cs",      "Czech"},
		{"csb",     "Kashubian"},
		{"cy",      "Welsh"},
		{"da",      "Danish"},
		{"de",      "German"},
		{"dv",      "Divehi; Dhivehi; Maldivian"},
		{"dz",      "Dzongkha"},
		{"el",      "Greek, Modern"},
		{"en",      "English"},
		{"eo",      "Esperanto"},
		{"es",      "Spanish; Castilian"},
		{"et",      "Estonian"},
		{"eu",      "Basque"},
		{"fa",      "Persian"},
		{"fi",      "Finnish"},
		{"fil",     "Filipino"},
		{"fo",      "Faroese"},
		{"fr",      "French"},
		{"fur",     "Friulian"},
		{"fy",      "Western Frisian"},
		{"ga",      "Irish"},
		{"gd",      "Gaelic; Scottish Gaelic"},
		{"gl",      "Galician"},
		{"gu",      "Gujarati"},
		{"gv",      "Manx"},
		{"ha",      "Hausa"},
		{"he",      "Hebrew"},
		{"hi",      "Hindi"},
		{"hr",      "Croatian"},
		{"hsb",     "Sorbian (Upper)"},
		{"ht",      "Haitian; Haitian Creole"},
		{"hu",      "Hungarian"},
		{"hy",      "Armenian"},
		{"ia",      "Interlingua (International Auxiliary Language Association)"},
		{"id",      "Indonesian"},
		{"ig",      "Igbo"},
		{"is",      "Icelandic"},
		{"it",      "Italian"},
		{"iu",      "Inuktitut"},
		{"ja",      "Japanese"},
		{"ka",      "Georgian"},
		{"kk",      "Kazakh"},
		{"kl",      "Kalaallisut (Greenlandic)"},
		{"km",      "Central Khmer"},
		{"kn",      "Kannada"},
		{"ko",      "Korean"},
		{"ks",      "Kashmiri"},
		{"ku",      "Kurdish"},
		{"kw",      "Cornish"},
		{"ky",      "Kirghiz; Kyrgyz"},
		{"la",      "Latin"},
		{"lg",      "Ganda"},
		{"li",      "Limburgan; Limburger; Limburgish"},
		{"lo",      "Lao"},
		{"lt",      "Lithuanian"},
		{"lv",      "Latvian"},
		{"mai",     "Maithili"},
		{"mg",      "Malagasy"},
		{"mi",      "Maori"},
		{"mk",      "Macedonian"},
		{"ml",      "Malayalam"},
		{"mn",      "Mongolian"},
		{"mr",      "Marathi"},
		{"ms",      "Malay"},
		{"mt",      "Maltese"},
		{"nan",     "Chinese, Min Nan"},
		{"nb",      "Bokmål, Norwegian; Norwegian Bokmål"},
		{"nds",     "German, Low"},
		{"ne",      "Nepali"},
		{"nl",      "Dutch; Flemish"},
		{"nn",      "Norwegian Nynorsk; Nynorsk, Norwegian"},
		{"nso",     "Sotho, Northern"},
		{"oc",      "Occitan (post 1500)"},
		{"om",      "Oromo"},
		{"or",      "Oriya"},
		{"pa",      "Panjabi; Punjabi"},
		{"pap",     "Papiamento"},
		{"pl",      "Polish"},
		{"pt",      "Portuguese"},
		{"ro",      "Romanian"},
		{"ru",      "Russian"},
		{"rw",      "Kinyarwanda"},
		{"sa",      "Sanskrit"},
		{"sc",      "Sardinian"},
		{"sd",      "Sindhi"},
		{"se",      "Northern Sami"},
		{"shs",     "Shuswap"},
		{"si",      "Sinhala; Sinhalese"},
		{"sk",      "Slovak"},
		{"sl",      "Slovenian"},
		{"so",      "Somali"},
		{"sq",      "Albanian"},
		{"sr",      "Serbian"},
		{"st",      "Sotho, Southern"},
		{"sv",      "Swedish"},
		{"ta",      "Tamil"},
		{"te",      "Telugu"},
		{"tg",      "Tajik"},
		{"th",      "Thai"},
		{"ti",      "Tigrinya"},
		{"tk",      "Turkmen"},
		{"tl",      "Tagalog"},
		{"tlh",     "Klingon"},
		{"tr",      "Turkish"},
		{"ts",      "Tsonga"},
		{"tt",      "Tatar"},
		{"ug",      "Uighur; Uyghur"},
		{"uk",      "Ukrainian"},
		{"ur",      "Urdu"},
		{"uz",      "Uzbek"},
		{"ve",      "Venda"},
		{"vi",      "Vietnamese"},
		{"wa",      "Walloon"},
		{"wo",      "Wolof"},
		{"xh",      "Xhosa"},
		{"yi",      "Yiddish"},
		{"yo",      "Yoruba"},
		{"zh",      "Chinese"},
		{"zh-hans", "Simplified Chinese"},
		{"zh-hant", "Traditional Chinese"},
		{"zu",      "Zulu"},
		{NULL,      NULL},
	};
	
	g_return_val_if_fail(language != NULL, "");
	
	for (i = 0; languages[i][0] != NULL; i++) {
		if (g_strcmp0(languages[i][0], language) == 0) {
			return languages[i][1];
		}
	}
	
	return language;
}

gchar *lark_tts_engine_format_voice_description(LarkTtsEngineVoice *voice)
{
	GString *langs;
	gchar *format, *langlist, *res;
	gint i;
	
	g_return_val_if_fail(voice != NULL, NULL);
	
	langs = g_string_new(NULL);
	
	for (i = 0; i < voice->languages->len; i++) {
		/*Format string*/
		switch (g_array_index(voice->genders, gint, i)) {
			case LARK_TTS_ENGINE_VOICE_GENDER_MALE:
				if (i == 0) {
					format = _("%s (Male)");
				} else {
					format = _(", %s (Male)");
				}
				break;
			case LARK_TTS_ENGINE_VOICE_GENDER_FEMALE:
				if (i == 0) {
					format = _("%s (Female)");
				} else {
					format = _(", %s (Female)");
				}
				break;
			default:
				if (i == 0) {
					format = "%s";
				} else {
					format = ", %s";
				}
				break;
		}
		/*Final string*/
		g_string_append_printf(langs, format, lark_tts_engine_find_language_description((gchar *)g_ptr_array_index(voice->languages, i)));
	}
	
	langlist = g_string_free(langs, FALSE);
	
	res = g_strdup_printf("%s\n<small>%s</small>", voice->name, langlist);
	
	g_free(langlist);
	
	return res;
}
